from __future__ import annotations

from tsbot import query

CLIENT_LIST_QUERY = query("clientlist").option("times")
CHANNEL_LIST_QUERY = query("channellist")
SERVER_GROUPS_QUERY = query("servergrouplist")



from __future__ import annotations

from contextlib import suppress
from typing import TYPE_CHECKING

from tsbot import plugin, query
from tsbot.exceptions import TSResponseError

from teamspeak_bot.common import CHANNEL_LIST_QUERY, CLIENT_LIST_QUERY
from teamspeak_bot.plugins import BasePluginConfig
from teamspeak_bot.utils import cache, find

if TYPE_CHECKING:
    from collections.abc import Callable

    from tsbot import TSBot, TSTask


DEFAULT_AFK_CHANNEL = "AFK"
DEFAULT_IDLE_TIME = 30 * 60  # 30 mins


class AFKMoverConfig(BasePluginConfig, total=False):
    afk_channel: str
    idle_time: float
    channel_whitelist: tuple[str, ...]
    channel_blacklist: tuple[str, ...]


DEFAULT_CONFIG = AFKMoverConfig(
    enabled=True,
)


def create_should_be_moved(
    max_idle_time: float,
    afk_channel_id: str,
    whitelist: tuple[str, ...] | None,
    blacklist: tuple[str, ...] | None,
) -> Callable[[dict[str, str]], bool]:
    def is_not_active(client: dict[str, str]) -> bool:
        return int(client.get("client_idle_time", 0)) > max_idle_time

    def not_in_afk_channel(client: dict[str, str]) -> bool:
        return client.get("cid", "") != afk_channel_id

    def is_not_query(client: dict[str, str]) -> bool:
        return client["client_type"] != "1"

    def not_in_blacklisted_channel(client: dict[str, str]) -> bool:
        return client["cid"] not in blacklist if blacklist else True

    def in_whitelisted_channel(client: dict[str, str]) -> bool:
        return client["cid"] in whitelist if whitelist else True

    checks = (
        is_not_active,
        not_in_afk_channel,
        is_not_query,
        not_in_blacklisted_channel,
        in_whitelisted_channel,
    )

    def should_be_moved(client: dict[str, str]) -> bool:
        return all(check(client) for check in checks)

    return should_be_moved


class AFKMover(plugin.TSPlugin):
    CHECK_INTERVAL = 60  # Check every minute

    def __init__(self, config: AFKMoverConfig) -> None:
        self.afk_channel = config.get("afk_channel", DEFAULT_AFK_CHANNEL)
        self.idle_time = config.get("idle_time", DEFAULT_IDLE_TIME) * 1000

        self.afk_channel_id: str = ""
        self.task: TSTask | None = None

        self.whitelisted_channels = config.get("channel_whitelist")
        self.blacklisted_channels = config.get("channel_blacklist")

        self.should_be_moved: Callable[[dict[str, str]], bool] = lambda c: False

    async def afk_mover_task(self, bot: TSBot) -> None:
        client_list = await cache.with_cache(
            bot.send, CLIENT_LIST_QUERY, max_ttl=self.CHECK_INTERVAL
        )

        to_be_moved = set(map(lambda c: c["clid"], filter(self.should_be_moved, client_list)))
        if not to_be_moved:
            return

        move_query = (
            query("clientmove")
            .params(cid=self.afk_channel_id)
            .param_block({"clid": clid} for clid in to_be_moved)
        )

        with suppress(TSResponseError):
            await bot.send(move_query)

    @plugin.once("connect")
    async def init_afk(self, bot: TSBot, ctx: None) -> None:
        self.afk_channel_id = await self.get_afk_channel(bot)
        self.should_be_moved = create_should_be_moved(
            self.idle_time,
            self.afk_channel_id,
            await self.get_whitelist_ids(bot),
            await self.get_blacklisted_ids(bot),
        )

    async def get_afk_channel(self, bot: TSBot) -> str:
        channel_list = await cache.with_cache(bot.send, CHANNEL_LIST_QUERY, max_ttl=60)

        channel_id = find.from_iterable(channel_list, self.afk_channel, "channel_name", "cid")
        if not channel_id:
            channel_id = await self.create_afk_channel(bot)

        return channel_id

    async def create_afk_channel(self, bot: TSBot) -> str:
        resp = await bot.send(
            query("channelcreate").params(
                channel_name=self.afk_channel, channel_flag_permanent=True
            )
        )

        return resp.first["cid"]

    async def get_whitelist_ids(self, bot: TSBot) -> tuple[str, ...] | None:
        if self.whitelisted_channels is None:
            return None

        channel_list = await cache.with_cache(bot.send, CHANNEL_LIST_QUERY, max_ttl=60)

        return tuple(
            c["cid"]
            for c in channel_list
            if any(wlc in c["channel_name"] for wlc in self.whitelisted_channels)
        )

    async def get_blacklisted_ids(self, bot: TSBot) -> tuple[str, ...] | None:
        if self.blacklisted_channels is None:
            return None

        channel_list = await cache.with_cache(bot.send, CHANNEL_LIST_QUERY, max_ttl=60)

        return tuple(
            c["cid"]
            for c in channel_list
            if any(blc in c["channel_name"] for blc in self.blacklisted_channels)
        )

    @plugin.on("connect")
    async def start_afk_mover(self, bot: TSBot, ctx: None) -> None:
        self.task = bot.register_every_task(
            self.CHECK_INTERVAL, self.afk_mover_task, name="AFKMover-Task"
        )

    @plugin.on("disconnect")
    async def stop_afk_mover(self, bot: TSBot, ctx: None) -> None:
        if self.task is not None:
            self.task = bot.remove_task(self.task)



from __future__ import annotations

import operator
from typing import TYPE_CHECKING, overload

if TYPE_CHECKING:
    from collections.abc import Iterable, Mapping


@overload
def from_iterable[K: str, V: str](
    search_list: Iterable[Mapping[K, V]],
    search_str: str,
    search_attr: K,
    result_attr: None = None,
    *,
    strict: bool = False,
) -> Mapping[K, V] | None: ...


@overload
def from_iterable[K: str, V: str](
    search_list: Iterable[Mapping[K, V]],
    search_str: str,
    search_attr: K,
    result_attr: K,
    *,
    strict: bool = False,
) -> V | None: ...


def from_iterable[K: str, V: str](
    search_list: Iterable[Mapping[K, V]],
    search_str: str,
    search_attr: K,
    result_attr: K | None = None,
    *,
    strict: bool = False,
) -> V | Mapping[K, V] | None:
    op = operator.eq if strict else operator.contains

    for item in search_list:
        if op(item[search_attr], search_str):
            return item[result_attr] if result_attr else item

    return None



from __future__ import annotations

import asyncio
import functools
from typing import TYPE_CHECKING

from tsbot import TSBot, TSCtx, TSTask, plugin, query
from tsbot.exceptions import TSCommandError

from teamspeak_bot.common import CHANNEL_LIST_QUERY, SERVER_GROUPS_QUERY
from teamspeak_bot.plugins import BasePluginConfig
from teamspeak_bot.utils import cache, checks, find, parsers, try_

if TYPE_CHECKING:
    from collections.abc import Mapping

DEFAULT_JAIL_CHANNEL = "Jail"
DEFAULT_INMATE_NAME = "Inmate"


class JailConfig(BasePluginConfig, total=False):
    jail_channel: str
    inmate_server_group_name: str
    can_jail_uids: tuple[str, ...]
    can_jail_server_groups: tuple[str, ...]
    strict_server_group_checking: bool


DEFAULT_CONFIG = JailConfig(
    enabled=False,
)

INMATE_SERVER_GROUP_PERMS = (
    {"permsid": "i_channel_join_power", "permvalue": -1, "permnegated": True},
    {"permsid": "i_group_needed_modify_power", "permvalue": 75},
    {"permsid": "i_group_needed_member_add_power", "permvalue": 75},
    {"permsid": "i_group_needed_member_remove_power", "permvalue": 75},
    {"permsid": "b_group_is_permanent", "permvalue": True},
    {"permsid": "b_client_is_sticky", "permvalue": True},
)


class JailPlugin(plugin.TSPlugin):
    FREE_QUERY = query("servergroupdelclient")
    JAIL_QUERY = query("servergroupaddclient")
    MOVE_QUERY = query("clientmove")

    def __init__(self, bot: TSBot, config: JailConfig) -> None:
        self.jail_channel = config.get("jail_channel", DEFAULT_JAIL_CHANNEL)
        self.inmate_name = config.get("inmate_server_group_name", DEFAULT_INMATE_NAME)

        self.jail_channel_id: str = ""
        self.inmate_server_group_id: str = ""

        self.jail_tasks: dict[str, TSTask] = {}

        can_jail_check = [
            checks.check_uids_and_server_groups(
                uids=config.get("can_jail_uids"),
                server_groups=config.get("can_jail_server_groups"),
                strict=config.get("strict_server_group_checking", False),
            )
        ]

        bot.register_command(
            "jail", self.jail, help_text="Jail a client for misbehaving", checks=can_jail_check
        )
        bot.register_command(
            "free", self.free, help_text="Free a client from jail", checks=can_jail_check
        )

    @plugin.once("connect")
    async def init_jail(self, bot: TSBot, ctx: None) -> None:
        self.jail_channel_id = await self.get_jail_channel_id(bot)
        self.inmate_server_group_id = await self.get_inmate_server_group_id(bot)
        await self.free_inmates_on_startup(bot)

    async def get_jail_channel_id(self, bot: TSBot) -> str:
        channel_list = await cache.with_cache(bot.send, CHANNEL_LIST_QUERY, max_ttl=60)

        channel_id = find.from_iterable(channel_list, self.jail_channel, "channel_name", "cid")
        if not channel_id:
            channel_id = await self.create_jail_channel(bot)

        return channel_id

    async def create_jail_channel(self, bot: TSBot) -> str:
        resp = await bot.send(
            query("channelcreate").params(
                channel_name=self.jail_channel, channel_flag_permanent=True
            )
        )

        return resp.first["cid"]

    async def get_inmate_server_group_id(self, bot: TSBot) -> str:
        server_groups_list = await cache.with_cache(bot.send, SERVER_GROUPS_QUERY, max_ttl=60)

        inmate_id = find.from_iterable(server_groups_list, self.inmate_name, "name", "sgid")
        if inmate_id is None:
            inmate_id = await self.create_inmate_server_group(bot)

        return inmate_id

    async def create_inmate_server_group(self, bot: TSBot) -> str:
        resp = await bot.send(query("servergroupadd").params(name=self.inmate_name))
        inmate_id = resp.first["cid"]

        await bot.send(
            query("servergroupaddperm")
            .params(sgid=inmate_id)
            .param_block(INMATE_SERVER_GROUP_PERMS)
        )

        return inmate_id

    async def free_inmates_on_startup(self, bot: TSBot) -> None:
        inmates = await bot.send(
            query("servergroupclientlist").params(sgid=self.inmate_server_group_id)
        )

        if not inmates.data:
            return

        await bot.send(
            self.FREE_QUERY.params(sgid=self.inmate_server_group_id).param_block(
                {"cldbid": c["cldbid"]} for c in inmates
            )
        )

    @plugin.on("cliententerview")
    async def handle_connect(self, bot: TSBot, ctx: TSCtx) -> None:
        if self.inmate_server_group_id in ctx["client_servergroups"].split(","):
            await bot.send(self.MOVE_QUERY.params(cid=self.jail_channel_id, clid=ctx["clid"]))

    async def jail(self, bot: TSBot, ctx: TSCtx, nickname: str, time: str) -> None:
        jail_time = try_.or_none(parsers.parse_time, time)
        if jail_time is None:
            raise TSCommandError("Invalid time format")

        client_list = await bot.send(query("clientlist").option("groups"))

        client = find.from_iterable(client_list, nickname, "client_nickname")
        if client is None:
            raise TSCommandError("No client found")

        if self.inmate_server_group_id in client["client_servergroups"].split(","):
            raise TSCommandError("Client already jailed")

        self.jail_tasks[client["client_database_id"]] = bot.register_task(
            functools.partial(
                self.jail_client_task,
                client=client,
                jail_time=jail_time,
            )
        )

    async def free(self, bot: TSBot, ctx: TSCtx, nickname: str) -> None:
        client_list = await bot.send(query("clientlist").option("groups"))

        client = find.from_iterable(client_list, nickname, "client_nickname")
        if client is None:
            raise TSCommandError("No client found")

        if self.inmate_server_group_id not in client["client_servergroups"].split(","):
            raise TSCommandError("Client not jailed")

        cldbid = client["client_database_id"]
        await bot.send(self.FREE_QUERY.params(sgid=self.inmate_server_group_id, cldbid=cldbid))

        if task := self.jail_tasks.get(cldbid):
            bot.remove_task(task)

    async def jail_client_task(self, bot: TSBot, client: Mapping[str, str], jail_time: int) -> None:
        clid = client["clid"]
        cldbid = client["client_database_id"]

        try:
            await bot.send(self.JAIL_QUERY.params(sgid=self.inmate_server_group_id, cldbid=cldbid))
            await bot.send(self.MOVE_QUERY.params(cid=self.jail_channel_id, clid=clid))

            await asyncio.sleep(jail_time)

            await bot.send(self.FREE_QUERY.params(sgid=self.inmate_server_group_id, cldbid=cldbid))

        finally:
            self.jail_tasks.pop(cldbid, None)



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Mapping


def deep_merge[T, K, V](from_: Mapping[K, V], to: T) -> T:
    for key, value in from_.items():
        if key in to and isinstance(to[key], dict) and isinstance(value, dict):  # type: ignore
            to[key] = deep_merge(to[key], value)  # type: ignore
        else:
            to[key] = value  # type: ignore

    return to



from __future__ import annotations

TIME_MAP = (
    ("h", 3600),
    ("m", 60),
    ("s", 1),
)


def parse_time(time: str) -> int:
    "Parse given time str (raw seconds or format XXhXXmXXs) into its seconds"

    if time.isnumeric():
        return int(time)

    unparsed = time
    parsed_time = 0

    for unit, multiplier in TIME_MAP:
        raw_time, sep, rest = unparsed.partition(unit)

        if not sep:
            continue

        parsed_time += int(raw_time) * multiplier

        if not rest:
            return parsed_time

        unparsed = rest

    raise ValueError("Incorrect time format")



from __future__ import annotations

import asyncio
import functools
from typing import TYPE_CHECKING, NamedTuple

import httpx
from result import Err, Ok, Result
from tsbot import plugin
from tsbot.exceptions import TSCommandError

from teamspeak_bot.plugins import BasePluginConfig

if TYPE_CHECKING:
    from tsbot import TSBot, TSCtx


class JokesPluginConfig(BasePluginConfig): ...


DEFAULT_CONFIG = JokesPluginConfig(
    enabled=True,
)


class Joke(NamedTuple):
    setup: str
    delivery: str


class JokesPlugin(plugin.TSPlugin):
    API_URL = "https://v2.jokeapi.dev/joke/Programming?type=twopart"

    async def get_a_joke(self) -> Result[Joke, str]:
        async with httpx.AsyncClient() as client:
            data = (await client.get(self.API_URL)).json()

            if data["error"]:
                return Err(data["message"])

            return Ok(Joke(data["setup"], data["delivery"]))

    @plugin.command("joke", help_text="Tells a programming joke")
    async def tell_a_joke(self, bot: TSBot, ctx: TSCtx) -> None:
        match await self.get_a_joke():
            case Err(error):
                raise TSCommandError(error)
            case Ok(joke):
                bot.register_task(functools.partial(self.tell_a_joke_task, ctx=ctx, joke=joke))

    async def tell_a_joke_task(self, bot: TSBot, ctx: TSCtx, joke: Joke) -> None:
        await bot.respond(ctx, joke.setup)
        await asyncio.sleep(3)
        await bot.respond(ctx, joke.delivery)



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Callable, Coroutine


def or_none[T, *Ts](func: Callable[[*Ts], T], *args: *Ts) -> T | None:
    try:
        return func(*args)
    except Exception:
        return None


def or_default[T, TD, *Ts](func: Callable[[*Ts], T], *args: *Ts, default: TD) -> T | TD:
    try:
        return func(*args)
    except Exception:
        return default


def or_call[T, TD, *Ts](
    func: Callable[[*Ts], T], *args: *Ts, on_error: Callable[[Exception], TD]
) -> T | TD:
    try:
        return func(*args)
    except Exception as e:
        return on_error(e)


async def async_or_call[T, TD](
    func: Coroutine[None, None, T], on_error: Callable[[Exception], TD]
) -> T | TD:
    try:
        return await func
    except Exception as e:
        return on_error(e)



from __future__ import annotations

import operator
from typing import TYPE_CHECKING, Concatenate

from tsbot.exceptions import TSPermissionError

from teamspeak_bot.utils import get

if TYPE_CHECKING:
    from collections.abc import Callable, Coroutine, Sequence

    from tsbot import TSBot, TSCtx


def has_group(
    groups: Sequence[str],
    client_groups: Sequence[dict[str, str]],
    *,
    strict: bool = False,
) -> bool:
    op = operator.eq if strict else operator.contains
    return any(op(cg["name"], g) for g in groups for cg in client_groups)


async def check_groups(
    bot: TSBot,
    ctx: TSCtx,
    groups: Sequence[str],
    *,
    strict: bool = False,
) -> bool:
    return has_group(groups, await get.client_server_groups(bot, ctx), strict=strict)


def check_uids(uids: Sequence[str], ctx: TSCtx) -> bool:
    return ctx.get("invokeruid") in uids


def check_uids_and_server_groups[**P](
    uids: tuple[str, ...] | None,
    server_groups: tuple[str, ...] | None,
    *,
    strict: bool = False,
) -> Callable[Concatenate[TSBot, TSCtx, P], Coroutine[None, None, None]]:
    async def is_allowed_to_run(bot: TSBot, ctx: TSCtx, *args: P.args, **kwargs: P.kwargs) -> None:
        if uids and check_uids(uids, ctx):
            return

        if server_groups and await check_groups(bot, ctx, server_groups, strict=strict):
            return

        raise TSPermissionError("Client not permitted to run this command")

    return is_allowed_to_run



from __future__ import annotations

import asyncio
import time
from typing import TYPE_CHECKING, Any, NamedTuple

if TYPE_CHECKING:
    from collections.abc import Callable, Coroutine

type CacheKey = tuple[Any, ...]
type Cache = dict[CacheKey, CacheRecord]


class CacheRecord(NamedTuple):
    timestamp: float
    record: Any


class _Cache:
    def __init__(self, max_ttl: float) -> None:
        self._cache: Cache = dict()
        self._max_ttl = max_ttl

    def get(self, key: CacheKey, max_ttl: float) -> Any | None:
        cache_hit = self._cache.get(key)
        if cache_hit is None:
            return None

        now = time.monotonic()
        if cache_hit.timestamp + max_ttl < now:
            return None

        return cache_hit.record

    def add(self, key: CacheKey, data: Any) -> None:
        self._cache[key] = CacheRecord(time.monotonic(), data)

    def purge(self) -> None:
        now = time.monotonic()

        to_be_deleted = tuple(
            key for key, item in self._cache.items() if item.timestamp + self._max_ttl < now
        )

        for key in to_be_deleted:
            del self._cache[key]


_cache = _Cache(max_ttl=60 * 10)
_cache_lock = asyncio.Lock()


async def with_cache[T, *Ts](
    func: Callable[[*Ts], Coroutine[None, None, T]], *args: *Ts, max_ttl: int
) -> T:
    async with _cache_lock:
        _cache.purge()

        if hit := _cache.get(args, max_ttl):
            return hit

        result = await func(*args)
        _cache.add(args, result)

        return result



from __future__ import annotations

import asyncio
import functools
from contextlib import suppress
from typing import TYPE_CHECKING

from tsbot import TSCtx, plugin, query
from tsbot.exceptions import TSCommandError, TSResponseError

from teamspeak_bot.plugins import BasePluginConfig
from teamspeak_bot.utils import formatters, parsers, try_

if TYPE_CHECKING:
    from tsbot import TSBot


DEFAULT_MAX_DELAY = 60 * 60


class NotifyConfig(BasePluginConfig, total=False):
    max_delay: int


DEFAULT_CONFIG = NotifyConfig(
    enabled=True,
)


class NotifyPlugin(plugin.TSPlugin):
    def __init__(self, config: NotifyConfig) -> None:
        self.max_delay = config.get("max_delay", DEFAULT_MAX_DELAY)

    @plugin.command(
        "notify",
        help_text="Pokes you after given amount of time in seconds or in XXh XXm XXs format",
    )
    async def notify_after(self, bot: TSBot, ctx: TSCtx, time: str, *message: str) -> None:
        delay = try_.or_none(parsers.parse_time, time)

        if delay is None:
            raise TSCommandError("Invalid time format")

        if self.max_delay and delay > self.max_delay:
            raise TSCommandError(f"Time must be under {formatters.seconds_to_time(self.max_delay)}")

        bot.register_task(
            functools.partial(
                self.notify_task,
                client_id=ctx["invokerid"],
                message=" ".join(message),
                delay=delay,
            )
        )

    async def notify_task(self, bot: TSBot, client_id: str, message: str, delay: int) -> None:
        await asyncio.sleep(delay)
        with suppress(TSResponseError):
            await bot.send(query("clientpoke").params(clid=client_id, msg=message))



from __future__ import annotations


def seconds_to_time(seconds: int) -> str:
    if seconds < 60:
        return f"{seconds}s"

    if seconds < 3600:
        m, s = divmod(seconds, 60)
        return f"{m}m{f' {s}s' if s else ''}"

    h, remaining = divmod(seconds, 3600)
    m, s = divmod(remaining, 60)
    return f"{h}h{f' {m}m' if m or s else ''}{f' {s}s' if s else ''}"
